/* Program version 0.92, released October 27, 2001.  Copyright Nathanael Hoyle 2001.  Please read both README and LICENSE files. */

#include <stdio.h>
#include <sys/ioctl.h>
#include <linux/cdrom.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

extern int errno;
short int waveout = 0;
void process_audio(int filedescriptor,int readtest,int retries,int start_track,int stop_track); /* Attempts to query further info from AUDIO cd's.  Takes open device handle descriptor, and whether to do full readtest, if so, then how many retries to attempt on a failed read. Also whether to test a single track. */

int main(int argc, char *argv[]){
	int drive,status,type;
	int readtest=0;
	int max_retries=0,start=0,stop=0;
	
	if(argc>1){
		if(!(strcmp(argv[1],"--reset"))){
			printf("Attempting to reset drive.\n");
			status=ioctl(drive,CDROMRESET);
			if(status==-1){
				if(errno==EACCES){
					perror("Permission denied error received.  Note that most systems will require root access to reset a drive.\n");
				}
				else{
					perror(strerror(errno));
				}
				perror("Error on resetting drive.  Assume reset failed.\n");
			}
			else{
				printf("Drive successfully reset.\n");
			}
			exit(0);
		}
		else if(!strcmp(argv[1],"--readtest")){
			readtest=1;
			if(argc>=3){
				max_retries=atoi(argv[2]);
				if(argc>=4){
					start=atoi(argv[3]);
					stop=start;
					if(argc>=5){
						stop=atoi(argv[4]);
					}
					if(argc==6){
						if(!strcmp(argv[5],"--waveout")){
							waveout=1;
						}
					}
				}
			}
		}	
		else if(!strcmp(argv[1],"--version")){
			printf("CD Drive Status and CD Analyzer / Reader by Nathanael Hoyle, version 0.92.  Copyright 2001.\n\n");
			exit(0);
		}
		else if(!strcmp(argv[1],"--help")){
			printf("CD Status 0.92 by Nathanael Hoyle (C) 2001\n--reset  reset the cd drive.  probably requires root.\n--readtest [failed read retries] [start track] [end track] [--waveout] does full frame by frame read of tracks on audio cd to test for errors.\nIf --waveout option specified (must use all --readtest options to use --waveout, sorry), tracks are written as .wav files.  See README\n--version  displays version info and exits\n--help  this usage info\n");
			exit(0);
		}
		else{
			printf("Invalid option selected. Syntax is: %s [--reset | --readtest [max retries] [start track] [end track] | --version | --help]\n",argv[0]);
			exit(1);
		}
	}
	printf("CD Drive Status and CD Analyzer / Reader by Nathanael Hoyle, version 0.92.  Copyright 2001.\n\n");
 
	drive = open("/dev/cdrom", O_RDONLY | O_NONBLOCK);
	if(drive==-1){
		perror("Error opening drive device.\n");
		exit(1);
	}
				

	status=ioctl(drive,CDROM_DRIVE_STATUS);
	if(status==-1){
		perror("Error getting drive status (while performing ioctl).\n");
		exit(1);
	}

	switch(status){
		case CDS_NO_INFO:
			printf("No information is available (i.e. this function is not implemented on the drive.\n");
			exit(0);
		case CDS_NO_DISC:
			printf("No disc is in the drive.\n");
			exit(0);
		case CDS_TRAY_OPEN:
			printf("The drive tray is open.  Please close it then retry.\n");
			exit(0);
		case CDS_DRIVE_NOT_READY:
			printf("The drive is not ready.\n");
			exit(0);
		case CDS_DISC_OK:
			printf("The drive reads disc ok.  Attempting to retrieve disc info.\n");
			break;
		default:
			printf("Error: Invalid status code returned: %d",status);
			exit(status);
	}

	type=ioctl(drive,CDROM_DISC_STATUS);

	if(type==-1){
		perror("Unexpected error reading disc information from drive.  Exiting.\n");
		exit(1);
	}

	switch(type){
		case CDS_NO_INFO:
			printf("Unable to read disc type info.  Function not implemented.\n");
			exit(0);
		case CDS_NO_DISC:
			printf("\"No disc in drive\" status returned.  Unusual since disc presence and drive readiness was just verified.\n");
			exit(0);
		case CDS_AUDIO:
			printf("Disc appears to be an audio cd.\n");
			process_audio(drive,readtest,max_retries,start,stop);
			break;
		case CDS_DATA_1:
			printf("Disc appears to be a DATA MODE 1 cd.\n");
			break;
		case CDS_DATA_2:
			printf("Disc appears to be a DATA MODE 2 cd.\n");
			break;
		case CDS_XA_2_1:
			printf("Disc appears to be of type XA 2 1\n");
			break;
		case CDS_XA_2_2:
			printf("Disc appears to be of type XA 2 2\n");
			break;
		case CDS_MIXED:
			printf("Disc appears to be MIXED MODE\nWill try to gather info on contained audio tracks.\n");
			process_audio(drive,readtest,max_retries,start,stop);
			break;
		default:
			printf("Unknown type returned: %d.  Possible error occured.",type);
	}

	
	close(drive);
	return 0;
}		

void process_audio(int drive,int readtest,int max_retries,int start,int stop){
	struct cdrom_tochdr cdtochdr;
	struct cdrom_tocentry cdtocentry;
	struct cdrom_read_audio cdra;
	struct cd_toc_info{
		int min,sec,frame_local,frame_global,data;
	} cdtocinfo[100];
	int status;
	int counter,counter2,readframes,scratch,retry;
	struct cdrom_volctrl cdvolume;
	FILE * audio_out;
	char filename[64];
	long int chunksize; /*for wav header*/
	long int scratchint; /*abused, misused, and reused int for all sorts of values that need to be written to file and hence need a variable*/
	short int scratchshort; /*same thing for 2-byte fields*/

	printf("Attempting to read table of contents.\n");
	status=ioctl(drive,CDROMREADTOCHDR,&cdtochdr);
	if(status==-1){
		perror("Unable to read TOC.\n");
		perror(strerror(errno));
		exit(1);
	}
	printf("Disc appears to have %d tracks. Querying for track info.\n",cdtochdr.cdth_trk1);

	for(counter=cdtochdr.cdth_trk0;counter<=cdtochdr.cdth_trk1;counter+=1){
		cdtocentry.cdte_format=CDROM_MSF;
		cdtocentry.cdte_track=counter;
		status=ioctl(drive,CDROMREADTOCENTRY,&cdtocentry);
		if(status==-1){
			printf("Error querying track info for track %d\n",counter);
			perror(strerror(errno));
		}
		else{
			cdtocinfo[counter].min=cdtocentry.cdte_addr.msf.minute;
			cdtocinfo[counter].sec=cdtocentry.cdte_addr.msf.second;
			cdtocinfo[counter].frame_local=cdtocentry.cdte_addr.msf.frame;
			cdtocinfo[counter].frame_global=cdtocinfo[counter].frame_local;
			(cdtocinfo[counter].frame_global)+=((cdtocinfo[counter].min)*CD_FRAMES*60);
			(cdtocinfo[counter].frame_global)+=((cdtocinfo[counter].sec)*CD_FRAMES);
			cdtocinfo[counter].frame_global-=CD_MSF_OFFSET;
			cdtocinfo[counter].data=(cdtocentry.cdte_ctrl&CDROM_DATA_TRACK);
			printf("Track %d starts at: minute: %d, second: %d, frame: %d\n",counter,cdtocinfo[counter].min, \
				cdtocinfo[counter].sec,cdtocinfo[counter].frame_global);
		}
	}

	printf("Gathering info on leadout track.\n");
	cdtocentry.cdte_format=CDROM_MSF;
	cdtocentry.cdte_track=CDROM_LEADOUT;
	
	status=ioctl(drive,CDROMREADTOCENTRY,&cdtocentry);
	if(status==-1){
		printf("Error querying track info for leadout track.\n");
		perror(strerror(errno));
	}
	else{
		cdtocinfo[0].min=cdtocentry.cdte_addr.msf.minute;
		cdtocinfo[0].sec=cdtocentry.cdte_addr.msf.second;
		cdtocinfo[0].frame_local=cdtocentry.cdte_addr.msf.frame;
		cdtocinfo[0].frame_global=cdtocinfo[0].frame_local;
		(cdtocinfo[0].frame_global)+=((cdtocinfo[0].min)*CD_FRAMES*60);
		(cdtocinfo[0].frame_global)+=((cdtocinfo[0].sec)*CD_FRAMES);
		cdtocinfo[0].frame_global-=CD_MSF_OFFSET;

		printf("Leadout track starts at: minute: %d, second: %d, frame: %d\n",cdtocinfo[0].min, \
			cdtocinfo[0].sec,cdtocinfo[0].frame_global);
	}

	printf("Querying drive volume settings:\n");
	status=ioctl(drive,CDROMVOLREAD,&cdvolume);
	if(status==-1){
		printf("Unable to read volume settings. Error: %s\n",strerror(errno));
	}
	else{
		printf("Volume settings for drive: Channel 1: %d, Channel 2: %d, Channel 3: %d, Channel 4: %d\n", \
			cdvolume.channel0,cdvolume.channel1,cdvolume.channel2,cdvolume.channel3);
	}
	
	if(readtest){
		printf("\n\nBeginning full read test.\n");
		if(waveout){
			printf("Writing output to wav files.\n");
		}
		cdra.buf=malloc(CD_FRAMESIZE_RAW);
		if(!start){
			start=cdtochdr.cdth_trk0;
		}
		if(!stop){
			stop=cdtochdr.cdth_trk1;
		}
			
		for(counter=start;counter<=stop;counter++){
			if(cdtocinfo[counter].data){
				printf("Skipping track %d.  (Data track).\n",counter);
				continue;
			}
			printf("Beginning track %d.  Progress:\n",counter);
			if(waveout){
				sprintf(filename,"track%02d.wav",counter);
			}
			else{
				sprintf(filename,"track%02d.raw",counter);
			}
			audio_out=fopen(filename,"w");
			cdra.nframes=1;	
			cdra.addr_format=CDROM_MSF;
			
			if(counter==cdtochdr.cdth_trk1){
				readframes=cdtocinfo[0].frame_global-cdtocinfo[counter].frame_global;
			
			}
			else{
				readframes=cdtocinfo[counter+1].frame_global-cdtocinfo[counter].frame_global;
			}
			cdra.addr.msf.minute=cdtocinfo[counter].min;
			cdra.addr.msf.second=cdtocinfo[counter].sec;
			cdra.addr.msf.frame=cdtocinfo[counter].frame_local;

			//begin of wav header output
			if(waveout){
				fprintf(audio_out,"RIFF");
				chunksize=readframes*CD_FRAMESIZE_RAW;
				chunksize+=36; /*don't ask*/
				fwrite((const void *)&chunksize,sizeof(long int),(size_t)1,audio_out);
				chunksize-=36; /*hey, I said don't ask*/
				fprintf(audio_out,"WAVEfmt "); /*overall wave format header, and format (fmt ) subchunk header in one*/
				scratchint=16;	/*subchunk 1 size*/
				fwrite((const void *)&scratchint,sizeof(long int),(size_t)1,audio_out);
				scratchshort=1; /*audio format.  PCM is 1 (linear quantization)*/
				fwrite((const void *)&scratchshort,sizeof(short int),(size_t)1,audio_out);
				scratchshort=2; /* number of channels. 2, as in stereo audio */
				fwrite((const void *)&scratchshort,sizeof(short int),(size_t)1,audio_out);
				scratchint=44100; /* sample rate */
				fwrite((const void *)&scratchint,sizeof(long int),(size_t)1,audio_out);
				scratchint=176400; /* ByteRate = SampleRate(44100) * NumChannels (2) * BitsPerSample(16) / 8 */
				fwrite((const void *)&scratchint,sizeof(long int),(size_t)1,audio_out);
				scratchshort=4; /* BlockAlign: bytes per full sample (2 channels * 16-bit samples = 4 bytes) */
				fwrite((const void *)&scratchshort,sizeof(short int),(size_t)1,audio_out);
				scratchshort=16; /* bits per sample.  normal signed 16 bit wave audio here */
				fwrite((const void *)&scratchshort,sizeof(short int),(size_t)1,audio_out);
				fprintf(audio_out,"data");
				fwrite((const void *)&chunksize,sizeof(long int),(size_t)1,audio_out);
			} //end of wav header stuff

			for(counter2=0;counter2<readframes;counter2++){
				status=ioctl(drive,CDROMREADAUDIO,&cdra);
				if(status==-1){
					if(errno==EOPNOTSUPP){
						printf("Error, the drive does not support reading audio in this mode.\nYou will not be able to use the --readtest option with this drive, sorry.");
						return;
					}
					printf("\nError trying to read track %d audio at minute: %d, second: %d, frame %d.",counter,cdra.addr.msf.minute,cdra.addr.msf.second,cdra.addr.msf.frame);
					printf(strerror(errno));
					printf("\n");
					fflush(stdout);
					for(retry=1;(retry<=max_retries)&&(status==-1);retry++){
						printf("Retrying...\n");
						status=ioctl(drive,CDROMREADAUDIO,&cdra);
					}
					if(status==-1){
						printf("All %d retries failed.  Expect tiny skip in audio.\n",max_retries);
					}
					else{
						printf("Succeeded on retry %d.\n",retry-1);
					}
				fflush(stdout);
				}
				if(status!=-1){
					fwrite((const void *)cdra.buf,(size_t)CD_FRAMESIZE_RAW,(size_t)1,audio_out);
				}
				cdra.addr.msf.frame++;
				if(!((cdra.addr.msf.frame)%CD_FRAMES)){
					cdra.addr.msf.frame=0;
					cdra.addr.msf.second++;
					if(cdra.addr.msf.second%60==0){
						cdra.addr.msf.second=0;
						cdra.addr.msf.minute++;
					}
				}
				if(counter2%(readframes/75)==0){
					printf("#");
					fflush(stdout);
				}
			}
			close(audio_out);
			printf("\nTrack %d completed.\n",counter);
		}
	free(cdra.buf);
	}
}
